require(MANTIS)
require(deSolve)
require(glue)
require(assertthat)
require(tidyverse)
require(magrittr)

season_beta <- function(time, beta, epsilon) beta * (1 + epsilon * sin(pi * time)^6)

required_sums <- function(Y, h, p, t) {
  h$totVirus <- sum(Y)
  # address seasonality in beta
  betai <- season_beta(t, p$beta, p$epsilon)
  # force of infection in z equations
  h$lambda <- betai * Y
  # amount of virus to prime cross-reactive antibody response
  for (i in 1:h$n_strains) {
    h$sum_shared_lambda[i] <- sum(betai[i] * Y[h$shared_epitopes[i,]])
  }
  # slower:
  # h$sum_shared_lambda <-
  #   (matrix(Y[h$shared_epitopes], nrow=nrow(h$shared_epitopes)) * betai) %>% rowSums()
  return(h)
}

# these are the definitions of the derivatives
dzdt <- function(z, mu, h) (1 - z) * h$lambda - mu * z
dwdt <- function(w, mu, h) {
  (1 - w) * h$sum_shared_lambda - mu * w
}
dydt <- function(y, z, w, gamma, sigma, mu, h) {
  ((1 - w) + (1 - gamma) * (w - z)) * h$lambda - sigma * y - mu * y
}
ddt <- function(x, h, p) {
  n_strains <- h$n_strains
  dzdt <- dzdt(z=x[1:n_strains], mu=p$mu, h=h)
  dwdt <- dwdt(w=x[(n_strains+1):(2*n_strains)], mu=p$mu, h=h)
  dydt <- dydt(z=x[1:n_strains],
               w=x[(n_strains+1):(2*n_strains)],
               y=x[(2*n_strains+1):length(x)],
               gamma=p$gamma, sigma=p$sigma, mu=p$mu, h=h)
  return(c(dzdt, dwdt, dydt))
}

# custom Runge-Kutta Cash-Karp parameterized solver
rkck <- function(x, dxdt, s, h, p) {
  b21 <- 0.2
  b31 <- 3/40
  b32 <- 9/40
  b41 <- 0.3
  b42 <- -0.9
  b43 <- 1.2
  b51 <- -11/54
  b52 <- 2.5
  b53 <- -70/27
  b54 <- 35/27
  b61 <- 1631/55296
  b62 <- 175/512
  b63 <- 575/13824
  b64 <- 44275/110592
  b65 <- 253/4096
  c1 <- 37/378
  c3 <- 250/621
  c4 <- 125/594
  c6 <- 512/1771
  xtemp <-  x + b21 * s * dxdt
  k2 <- ddt(xtemp, h, p)
  xtemp <-  x + s * (b31 * dxdt + b32 * k2)
  k3 <- ddt(xtemp, h, p)
  xtemp <- x + s * (b41 * dxdt + b42 * k2 + b43 * k3)
  k4 <- ddt(xtemp, h, p)
  xtemp <- x + s * (b51 * dxdt + b52 * k2 + b53 * k3 + b54 * k4)
  k5 <- ddt(xtemp, h, p)
  xtemp <- x + s * (b61 * dxdt + b62 * k2 + b63 * k3 + b64 * k4 + b65 * k5)
  k6 <- ddt(xtemp, h, p)
  return(x + s * (c1 * dxdt + c3 * k3 + c4 * k4 + c6 * k6))
}

loci_epitope <- function(n_strains, nEqs, nShEp, shared_epitopes, initCondY,
                         tMax, obsInt, tInt, beta, gamma, mu, sigma, epsilon,
                         initCondZ, initCondW) {
  x <- vector("numeric", nEqs)
  dxdt <- vector("numeric", nEqs)
  n_req_obs <- floor(tMax/obsInt) + 1
  t_vector <- vector("numeric", n_req_obs)
  y_array <- matrix(nrow=n_req_obs, ncol=nEqs)
  t <- 0
  timestep <- 1
  #assign parameter values to 'p'
  p <- list()
  p$beta <- beta
  p$gamma <- gamma
  p$mu <- mu
  p$sigma <- sigma
  p$epsilon <- epsilon
  #assign housekeeping values to 'h'
  h <- list()
  h$n_strains <- n_strains #number of strains
  h$nShEp <- nShEp #number of strains with which you share epitopes [including yourself]
  h$nEqs <- nEqs #total number of equations
  h$totVirus <- 0 #total amount of virus
  h$shared_epitopes <- shared_epitopes #shared epitope matrix
  # sum of virus that shares epitopes with each strain
  # [i.e. force of infection for each virus in w equation]
  h$sum_shared_lambda <- vector("numeric", h$n_strains)
  h$lambda <- vector("numeric", h$n_strains) #force of infection for each virus in z equation
  #set up intial conditions
  x <- c(initCondZ, initCondW, initCondY)
  #record intial state
  t_vector[timestep] <- t #record starting time
  y_array[timestep,] <- x #record intial conditions
  timestep <- timestep + 1 #move observation marker to next desired time
  #run through the simulation
  progstep <- 1
  progsteps_width <- 100
  while (t <= tMax) {
    t <- t + tInt
    #use timestep (obs) to give a proxy of computing progress
    # if ((progstep %% ((tMax / tInt) %/% progsteps_width)) == 0) {
    #     cat("\r[", str_c(str_c(rep("#", round(progsteps_width*t/tMax)), collapse=""),
    #                      str_c(rep(" ", round(progsteps_width*(tMax - t)/tMax)), collapse="")),
    #         "]", round(100*t/tMax), "%")}
    #do the necessary sums for this timestep
    # browser()
    h <- required_sums(x[(2*n_strains+1):length(x)], h, p, t)
    dxdt <- ddt(x, h, p) #calculate the derivatives
    x <- rkck(x, dxdt, tInt, h, p)
    #if have reached/passed next observation point, then record results ...
    if (t >= (timestep - 1) * obsInt) {
      t_vector[timestep] <- t
      y_array[timestep,] <- x
      timestep <- timestep + 1
    }
    progstep <- progstep + 1
  }
  # cat("\n")
  return(list(time=t_vector, Y=y_array))
}

run_mantis <- function(epiStruc, tMax, tObsPer, tStep, initCondY, beta, gamma,
                       sigma, mu, epsilon=0, initCondZ=NA, initCondW=NA) {
  assert_that(length(epiStruc) > 1, length(which(epiStruc < 0)) == 0)

  n.loci <- length(epiStruc)
  strainmx <- as.data.frame(expand.grid(lapply(epiStruc, seq, from=1, by=1)))
  n_strains <- dim(strainmx)[1]
  colnames(strainmx) <- paste("locus", 1:n.loci, sep="")
  rownames(strainmx) <- NULL

  assert_that(is.numeric(tMax), tMax > 0)
  assert_that(tObsPer %% 1 == 0, tObsPer >= 1,
              msg="tObsPer should be an integer greater than 1")
  assert_that(is.numeric(tStep), tStep > 0)
  assert_that(length(initCondY) > 0, length(which(initCondY < 0)) == 0,
              sum(initCondY) <= 1, length(initCondY) == n_strains)
  assert_that(is.numeric(beta), all(beta > 0))
  assert_that(is.numeric(gamma), gamma >= 0, gamma <= 1)
  assert_that(is.numeric(epsilon), epsilon >= 0)
  assert_that(is.numeric(sigma), sigma >= 0)
  assert_that(is.numeric(mu), mu >= 0)
  assert_that(length(beta) == 1 | length(beta) == n_strains,
              msg=glue("beta has to be a number (in which case all strains have the \\
                same beta) or a numeric array of length n_strains."))

  if (are_equal(initCondZ, NA)) {
    initCondZ <- rep(0, n_strains)
  } else assert_that(is.numeric(initCondZ), length(initCondZ) == n_strains)

  if (are_equal(initCondW, NA)) {
    initCondW <- rep(0, n_strains)
  } else assert_that(is.numeric(initCondW), length(initCondW) == n_strains)

  if (length(beta) == 1) beta <- rep(beta, n_strains)

  for (i in 1:n_strains) {
    temp <- which(strainmx[, 1] == strainmx[i, 1])
    for (j in 2:n.loci) {
      temp <- union(temp, which(strainmx[, j] == strainmx[i, j]))
    }
    if (i == 1) shared_epitopes <- matrix(0, n_strains, length(temp))
    shared_epitopes[i,] <- sort(temp)
  }

  nShEp <- dim(shared_epitopes)[2]
  nEqs <- 3 * n_strains
  tObsStep <- tObsPer * tStep
  # cat("Running MANTIS, please wait...\n")
  X <- loci_epitope(n_strains, nEqs, nShEp, shared_epitopes,
                    initCondY, tMax, tObsStep, tStep, beta, gamma, mu, sigma,
                    epsilon, initCondZ, initCondW)
  X <- c(X, list(shared_epitopes), list(strainmx))
  return(X)
}

# struct <- c(2,2)
# microbenchmark::microbenchmark(
# mantis_output_0 <- runMANTIS(struct,
#           tMax=1000, tObsPer=1, tStep=0.00001,
#           initCondY=runif(prod(struct)) %>% divide_by(sum(.)),
#           beta=1600, gamma=0.7, sigma=400, mu=0.02)#,
#     run_mantis(struct,
#                tMax=100, tObsPer=10, tStep=0.005,
#                initCondY=runif(prod(struct)) %>% divide_by(sum(.)),
#                beta=300, gamma=0.5, sigma=1/2*365, mu=1/100),
#     times=20)
## about 20x slower

# set.seed(0)
# init <- runif(prod(struct)) %>% divi pde_by(sum(.))
# mantis_output_0 <- run_mantis(struct,
#            tMax=100, tObsPer=10, tStep=0.005,
#            initCondY=init, initCondW=init, initCondZ=init,
#            beta=300, gamma=0.5, sigma=1/2*365, mu=1/100)
# library(MANTIS)
# set.seed(0)
# mantis_output_old <- runMANTIS(struct,
#                             tMax=100, tObsPer=10, tStep=0.005,
#                             initCondY=runif(prod(struct)) %>% divide_by(sum(.)),
#                             beta=300, gamma=0.5, sigma=1/2*365, mu=1/100)
# near(mantis_output$Y, mantis_output_old$Y) %>% all()

# timeseries_0 <- mantis_output_0$Y %>%
#     as.data.frame() %>%
#     set_names(list(lapply(struct, . %>% `:`(1,.)), list(c("_z", "_w", "_y"))) %>%
#                   unlist(recursive=FALSE) %>%
#                   expand.grid() %>%
#                   unite("strain", everything(), sep="/") %>%
#                   unlist()) %>%
#     mutate(time = mantis_output_0$T) %>%
#     gather("strain", "prevalence", -time) %>%
#     separate(strain, c("strain", "variable"), sep="/_") %>%
#     mutate(variable = factor(variable, levels=c("z", "w", "y"),
#                              labels=c("specific immunity",
#                                       "cross-reactive immunity",
#                                       "currently infectious")))
# ggplot(timeseries_0 %>% filter(time > 100)) +
#     aes(colour=strain, y=prevalence, x=time) +
#     geom_line() +
#     facet_grid(variable~strain, scales="free_y") +
#     scale_y_continuous(limits=c(0,1)) +
#     theme_bw()
