using DelimitedFiles
using ShiftedArrays
using Distributed

# strainstruct = strainstructure = [2 2]; mu = 0.02; rnaught = 4; sigma = 50; gamma = 0.75; maxtime = tmax = 10000.0; chi = 0.0; npop=1; beta = 200; nu = 0; epsilon = 0;

# while nprocs() < length(Sys.cpu_info())
    addprocs(7)
# end
# print(nprocs())

@everywhere begin
    Base.include(Main, "multipop_MANTIS.jl")
    using ShiftedArrays
    function count_unique_local_minima(vector, digits=nothing)
        # -1 = extinct, 0 = stable, 1 = unconverged, >1 = chaos/cycles
        vector = isnothing(digits) ? vector : round.(vector, digits=digits)
        # if the last element is 0, the strain is extinct, regardless of other dynamics
        if vector[end] == 0 return(-1) end
        # pick out the local minima (use isless bc handles missing values introduced by lead/lag
        # silently: missing > all numbers)
        minima = vector[isless.(vector, ShiftedArrays.lag(vector)) .& isless.(vector, ShiftedArrays.lead(vector))]
        number_unique_minima = length(unique(minima))
        # if there are no minima, the strain is constant or has a period proportional to the stepsize (unlikely)
        # if only one minima, the vector is monotonic (implies failure to converge)
        if number_unique_minima <= 1 return(number_unique_minima) end
        # now check for negative monotonicity in cyclical dynamics => assume going extinct
        if all(((minima - ShiftedArrays.lag(minima)) .<= 0)[2:end]) return(-1) end
        # if positive monotonicity in minima along with negative monotonicity in maxima, then call it stable (just not yet converged)
        if all(((minima - ShiftedArrays.lag(minima)) .>= 0)[2:end])
            maxima = vector[isless.(ShiftedArrays.lag(vector), vector) .& isless.(ShiftedArrays.lead(vector), vector)]
            if all(((maxima - ShiftedArrays.lag(maxima)) .<= 0)[2:end]) return(0) else return(1) end
        end
        # if not caught by any of the earlier conditions, then the dynamics are chaotic/cyclical
        return(number_unique_minima)
    end
    function clean_eps_timeseries_output(tmp, params, npopulations, nstrains)
        ntimepoints = size(tmp, 1)
        tmp = DataFrames.stack(tmp, 1:size(tmp, 2) - 1, size(tmp, 2))
        tmp.variable = vcat([[x for y in 1:2*ntimepoints] for x in 1:nstrains]...)
        tmp.population = vcat([vcat(fill(1, ntimepoints), fill(2, ntimepoints)) for x in 1:nstrains]...)
        tmp = DataFrames.by(tmp, [:variable, :population], :value => x -> count_unique_local_minima(x))
        tmp.gamma = vcat(fill(params[1:2], nstrains)...)
        tmp.sigma = vcat(fill(params[3:4], nstrains)...)
        tmp.R0 = fill(params[5], nstrains * npopulations)
        return(tmp)
    end
    function clean_eps1p_timeseries_output(tmp, params, nstrains)
        ntimepoints = size(tmp, 1)
        tmp = DataFrames.stack(tmp, 1:size(tmp, 2) - 1, size(tmp, 2))
        tmp.variable = vcat([[x for y in 1:ntimepoints] for x in 1:nstrains]...)
        tmp = DataFrames.by(tmp, [:variable], :value => x -> count_unique_local_minima(x),
                                              #:value => x -> count_unique_local_minima(x, 12),
                                              #:value => x -> count_unique_local_minima(x, 10),
                                              :value => x -> count_unique_local_minima(x, 8))
        tmp[!, :gamma] .= params[1]
        tmp[!, :sigma] .= params[2]
        tmp[!, :R0] .= params[3]
        tmp[!, :mu] .= params[4]
        return(tmp)
    end
    function clean_vmr_timeseries_output(tmp, params, npopulations, nstrains)
        ntimepoints = size(tmp, 1)
        tmp = DataFrames.stack(tmp, 1:size(tmp, 2) - 1, size(tmp, 2))
        tmp.variable = vcat([[x for y in 1:2*ntimepoints] for x in 1:nstrains]...)
        tmp.population = vcat([vcat(fill(1, ntimepoints), fill(2, ntimepoints)) for x in 1:nstrains]...)
        tmp = DataFrames.by(tmp, [:variable, :population], :value => x -> count_unique_local_minima(x),
                                                           :value => x -> count_unique_local_minima(x, 12),
                                                           :value => x -> count_unique_local_minima(x, 10),
                                                           :value => x -> count_unique_local_minima(x, 8),
                                                           :value => x -> count_unique_local_minima(x, 6),
                                                           :value => x -> count_unique_local_minima(x, 4))
        tmp.chi = fill(params[1], nstrains * npopulations)
        tmp.gamma = fill(params[2], nstrains * npopulations)
        tmp.sigma = fill(params[3], nstrains * npopulations)
        tmp.R0 = vcat(fill(params[4:5], nstrains)...)
        return(tmp)
    end
end

function explore_parameter_space(params, start_ind, stop_ind;
                                 strainstruct = [2 2],
                                 chi           = 0.05,
                                 maxtime       = 10000.0,
                                 gamma         = 0.01:0.33:0.99,
                                 sigma         = 0.05 .* (10 .^ (1:1:4)),
                                 rnaught       = [2,4,6],
                                 mu            = 0.02)
    if ismissing(params)
        params = [[γ1, γ2, σ1, σ2, R0] for γ1 in gamma for σ1 in sigma for γ2 in gamma for σ2 in sigma for R0 in rnaught]
        outfile = nothing
    else
        outfile = "../Results/SSE/$strainstruct;$chi;$maxtime;$start_ind;$stop_ind;$mu.csv"
    end
    nstrains = trunc(Int, prod(strainstruct))
    npopulations = 2
    if size(chi) != (2, 2)
         if length(chi) == 1
             chi = [-chi chi; 0 0]
         else error("Invalid chi value: this function assumes two populations") end
     end
    results = pmap(x -> clean_eps_timeseries_output(
                            runMANTIS(strainstructure=strainstruct,
                                      tmax=maxtime,
                                      tstep=trunc(Int, 0.8*maxtime):maxtime,
                                      # R0 = beta / (sigma + mu + disease-induced mortality)
                                      # note here we assume 0 disease-induced mortality
                                      beta=x[5] .* (x[3:4] + mu), gamma=x[1:2], sigma=x[3:4],
                                      chi=chi, mu=mu)["timeseries"][:,vcat((npopulations * nstrains + 1):(2 * npopulations * nstrains), end)], x, npopulations, nstrains),
                   params)
    if ismissing(outfile)
        return(convert(Matrix, vcat(results...)))
    else
        writedlm(outfile, convert(Matrix, vcat(results...)), ',')
        return(0)
    end
end

function explore_parameter_space_one_pop(params, start_ind, stop_ind;
                                         strainstruct = [2 2],
                                         maxtime      = 10000.0,
                                         gamma        = 0.02:0.04:0.98,
                                         sigma        = 10.0 .^ (-0.83:0.17:3.25),
                                         mu           = [0.02 0.05 0.1],
                                         rnaught      = [4/3 2 4])
    if ismissing(params)
        params = [[γ, σ, R0, μ] for γ in gamma for σ in sigma for R0 in rnaught for μ in mu]
        outfile = nothing
    else
        outfile = "../Results/SSE1p/$strainstruct;$maxtime;$start_ind;$stop_ind.csv"
    end
    nstrains = trunc(Int, prod(strainstruct))
    results = pmap(x -> clean_eps1p_timeseries_output(
                            runMANTIS(strainstructure=strainstruct,
                                      tmax=maxtime,
                                      tstep=trunc(Int, 0.8*maxtime):maxtime,
                                      beta=x[3] * (x[2] + x[4]), gamma=x[1], sigma=x[2],
                                      chi=0.0, mu=x[4])["timeseries"][:,vcat((nstrains + 1):(2 * nstrains), end)], x, nstrains),
                   params)
    if ismissing(outfile)
        return(convert(Matrix, vcat(results...)))
    else
        writedlm(outfile, convert(Matrix, vcat(results...)), ',')
        return(0)
    end
end

function variable_movement_rate(params, start_ind, stop_ind;
                                strainstruct = [2 2],
                                chi          = 10.0 .^ -(1:6),
                                maxtime      = 1000000.0,
                                gamma        = [0.66 0.75],
                                sigma        = 7.5,
                                rnaught      = [4/3, 2],
                                mu           = 0.15)
    if ismissing(params)
        params = [[χ γ σ R01 R02] for χ in chi for γ in gamma for σ in sigma for R01 in rnaught for R02 in rnaught for rep in 1:10]
        params = params[[x[4] != x[5] for x in params]]
        outfile = nothing
    else
        outfile = "../Results/VMR/$strainstruct;$maxtime;$start_ind;$stop_ind;$mu.csv"
    end

    nstrains = trunc(Int, prod(strainstruct))
    npopulations = 2
    results = pmap(x -> clean_vmr_timeseries_output(
                            runMANTIS(strainstructure=strainstruct,
                                      tmax=maxtime,
                                      tstep=trunc(Int, 0.9*maxtime):maxtime,
                                      beta=x[4:5] .* (x[3] .+ mu), gamma=x[2], sigma=x[3],
                                      chi=[-x[1] x[1]; 0 0], mu=mu .- [x[1]; 0])["timeseries"][:,vcat((npopulations * nstrains + 1):(2 * npopulations * nstrains), end)], x, npopulations, nstrains),
                   params)
    if isnothing(outfile)
        return(convert(Matrix, vcat(results...)))
    else
        writedlm(outfile, convert(Matrix, vcat(results...)), ',')
        return(0)
    end
end
