\name{MANTIS-package}
\alias{MANTIS-package}
\alias{MANTIS}
\docType{package}
\title{
Multilocus ANTIgenic Simulator
}
\description{
	This package can be used to simulate the population dynamics of antigenically diverse pathogens according to the multilocus antigenic framework by Gupta et al. (see reference 1). Its functions allow the user to generate host-pathogen time-series and analyse critical dynamic patterns within; both can be exported into comma-separated values files (CSV) and PDF plots. \cr\cr
	MANTIS is developed by the Evolutionary Ecology of Infectious Disease (EEID) research group at the Department of Zoology of the University of Oxford.\cr\cr
	\bold{Summary of the framework:}
		Transmission follows some basic assumptions of a classic susceptible-infectious-recovered (SIR) epidemiological framework. Hosts are segregated into 3 pertinent classes, \code{Z}, \code{W}, and \code{Y}. For each possible pathogen strain \code{i} in the system, individuals with cross-reactive immune responses are represented by \code{Wi}, infectious by \code{Yi} and with specific responses by \code{Zi} (for the corresponding ordinary-differential equations the reader should refer to reference 2). A pathogen strain is defined by the genetic loci that encode relevant antigens, against which immune responses act, rather than by its entire genome. For example, the notation \code{{3,4,10}}, would indicate a system with three loci \code{i,k,j} and \code{N_i=3}, \code{N_k=4} and \code{N_j=10} alleles, respectively (for any given structure, the total number of possible strains is \code{N_i * N_k * ... * N_n}. Strains that do not share alleles do not interfere with each others' success as mediated by host immune responses. The degree to which infection with a given strain limits the ability of another strain (sharing alleles) to infect the same host is defined by a cross-immunity parameter \code{gamma}. If \code{gamma=0} the strains do not induce cross-protective responses. If \code{gamma=1} there is complete cross-protection. Otherwise, immunity to a given strain only reduces the probability of transmission of another by a factor \code{1-gamma}. 
	\cr\cr
	\bold{Methodology:}
		These simple but realistic biological assumptions can be expressed by a system of ordinary-differential equations (ODE) representing the changes in the abundances of each of the strains over time in a genetically and demographically homogeneous host population. Here, a range-kutta algorithm with fixed-step size is used to solve an ODE system, thus numerically simulating the above described epidemiological framework.
	\cr\cr
	\bold{This version:}
		MANTIS V3.0 (cryptonym: "red egg")
	\cr\cr
	\bold{Referencing:}
		Please reference the MANTIS research article: 'MANTIS: an R package that simulates multilocus models of pathogen evolution. BMC Bioinformatics. 2015 May 28;16:176. doi: 10.1186/s12859-015-0598-9.'
	\cr\cr	
	\bold{Installation (source):}
		Open an R session. Make sure the session is open in the same folder as the MANTIS package file. Run \code{install.packages("XXX",repos=NULL,type="src")}, where XXX is the package file name. Note: MANTIS requires the preinstallation of Rcpp (see \url{http://cran.r-project.org/web/packages/Rcpp/index.html}); and, installing source packages on machines with the Windows operating system may require the preinstallation of Rtools (see \url{http://cran.r-project.org/bin/windows/Rtools/}).
	\cr\cr
	\bold{Installation (repository):}
		MANTIS is not available in official R repositories.
	\cr\cr
	\bold{Getting Started:}
		For a general example, see the 'examples' section on this document.
	\cr\cr
	\bold{Updates:} 
		Updates are planned to be submitted to official repositories and will also be available at the Evolutionary Ecology of Infectious Disease (EEID) group's webpage (see "See Also" for links)
	
}
\references{
  (1) Chaos, persistence, and evolution of strain structure in antigenically diverse infectious agents. Gupta S, Ferguson N, Anderson R. Science. 1998 May 8;280(5365):912-5. \cr\cr

  (2) The antigenic evolution of influenza: drift or thrift? Wikramaratna PS, Sandeman M, Recker M, Gupta S. Philos Trans R Soc Lond B Biol Sci. 2013 Feb 4;368(1614):20120200. \cr\cr

  (3) Transient cross-reactive immune responses can orchestrate antigenic variation in malaria. Recker M, Nee S, Bull PC, Kinyanjui S, Marsh K, Newbold C, Gupta S. Nature. 2004 Jun 3;429(6991):555-8. \cr\cr

  (4) Role of stochastic processes in maintaining discrete strain structure in antigenically diverse pathogen populations. Buckee CO, Recker M, Watkins ER, Gupta S. Proc Natl Acad Sci U S A. 2011 Sep 13;108(37):15504-9. \cr\cr

  (5) The generation of influenza outbreaks by a network of host immune responses against a limited set of antigenic types. Recker M, Pybus OG, Nee S, Gupta S. Proc Natl Acad Sci U S A. 2007 104, 7711-7716. \cr\cr

}
\keyword{ package }
\seealso{

Evolutionary Ecology of Infectious Disease (EEID) group\cr
\url{www.eeid.ox.ac.uk} \cr

EEID on Twitter, follow us, get in touch!\cr
\url{www.twitter.com/EEID_oxford} \cr

Department of Zoology of the University of Oxford\cr
\url{www.zoo.ox.ac.uk} \cr


}
\examples{

library(MANTIS)

#define epitope structure, with 2 loci
#with 4 and 2 alleles each (8 strains)
epiStruc<- c(4,2) 
#get number of strains in the system
nStrains<- extractNumberStrains(epiStruc) 
#time at which the solver will stop
tMax<- 1000
#solver step
tInt<- 0.005 
#record solution every tObsPer steps
tObsPer<- 100 
#random initial conditions for infected
initCondY <- runif(nStrains, 0, 1)*1e-4 

#parameter values here defined by year
#thus output time scale will be years
beta<-  292
sigma<-  (1/5)*365 #5 days infectious period
mu<-  1/50 #50 years of life-span

#running model with weak cross-immunity
gamma<- 0.1
simdata<- runMANTIS(epiStruc, tMax, tObsPer, 
	tInt, initCondY, beta, gamma, sigma, mu)

#plotting output with weak cross-immunity to the screen
plotY(simdata, xiObs=0.1, xfObs=0.2)

#plotting output with weak cross-immunity to a PDF file
plotY(simdata, fileout="weak.ts.pdf", xiObs=0.1, xfObs=0.2)

#plotting diversity to PDF files
plotYDiversity(simdata,fileout="diveristyY.pdf",
	method="shannon",xiObs=0.5, xfObs=0.8)

#export diversity of Z class into a text file
exportZDiversity(simdata, fileout="diversityZ.csv")

#calculate single-strain dominance
epsilon<-  calcSingleStrainDominance(simdata)	
cat(paste("Single-strain dominance is epsilon=", epsilon, "\n"))

#running model with intermediate cross-immunity
gamma<- 0.6
simdata<- runMANTIS(epiStruc, tMax, tObsPer, 
	tInt, initCondY, beta, gamma, sigma, mu)

#select strains 1 and 2
subSetStrains<- c(1,2)
#plot highlighting a subset of strains to file
plotY(simdata, fileout="highlighting.strains_1_2.pdf", colours=c("blue","grey"),
 markStrains=subSetStrains, xiObs=0.5, xfObs=0.53)

#get strains that share epitopes with strain 1, including 1
strainsShareWith1<- withSharedEpitopes(simdata,1) 
#plot highlighting a subset of strains
plotY(simdata, colours=c("orange","black"), markStrains=strainsShareWith1,
	xiObs=0.5, xfObs=0.53)
#plotting diversity to PDF files
plotYDiversity(simdata, method="shannon", xiObs=0.5, xfObs=0.53)

#exporting output into a CSV text file
exportTS(simdata, fileout="intermediate.all.strains.csv")

#exporting shared epitopes matrix into a CSV text file
exportSharedEpitopes(simdata, fileout="intermediate.shared.epitopes.csv")

#plot relative prevalence of the strains in time
plotYRelPrevalence(simdata, fileout="relative.prevalence.plot.pdf", 
	xiObs=0.5, xfObs=0.54, method="host")

#measure single-strain dominance from a range of cross-immunity values, gamma
gammas<- seq(0,1,length.out=10)

#measure within a small antigenic space
epiStruc<- c(2,2)
initCondY <- runif(extractNumberStrains(epiStruc), 0, 1)*1e-4 
SSD_sAntiSpace<- measureSSDfromGammaRange(gammas, "host", epiStruc, tMax, 
	tObsPer, tInt, tStep, initCondY, beta, sigma, mu)

#measure within a medium sized antigenic space
epiStruc<- c(4,2)
initCondY <- runif(extractNumberStrains(epiStruc), 0, 1)*1e-4 
SSD_mAntiSpace<- measureSSDfromGammaRange(gammas, "host", epiStruc, tMax, 
	tObsPer, tInt, tStep, initCondY, beta, sigma, mu)

#measure within a larger sized antigenic space
epiStruc<- c(6,2)
initCondY <- runif(extractNumberStrains(epiStruc), 0, 1)*1e-4 
SSD_lAntiSpace<- measureSSDfromGammaRange(gammas, "host", epiStruc, tMax, 
	tObsPer, tInt, tStep, initCondY, beta, sigma, mu)

#plot SSD measures agains gamma
colours=c("blue3","gold","tomato3")
pdf("SSDs_various_gammas_systems.pdf", width=5, height=4, bg="white")
plot(gammas, SSD_sAntiSpace, t='b', col=colours[1], lwd=2, ylab="SSD")
lines(gammas, SSD_mAntiSpace, t='b', col=colours[2], lwd=2)
lines(gammas, SSD_lAntiSpace, t='b', col=colours[3], lwd=2)
legend("topleft", legend=c("small antigenic space","med. antigenic space"
	,"big antigenic space"),col=colours, lwd=2, cex=0.8)
a<- dev.off()

###will define a bigger antigenic space to have a look at antigenic dynamics

#define epitope structure
epiStruc<- c(5,5,5) 
#get number of strains in the system
nStrains<- extractNumberStrains(epiStruc) 
#time at which the solver will stop
tMax<- 500
#solver step
tInt<- 0.005 
#record solution every tObsPer steps
tObsPer<- 100 
#random initial conditions for infected
initCondY <- runif(nStrains, 0, 1)*1e-9 

#parameter values here defined by year
#thus output time scale will be years
beta<- 292
sigma<-  (1/5)*365 #5 days infectious period
mu<-  1/50 #50 years of life-span

#running model with weak cross-immunity
gamma<- 0.75
simdata<- runMANTIS(epiStruc, tMax, tObsPer, 
	tInt, initCondY, beta, gamma, sigma, mu)

#plotting antigenic dynamics
plotAntigenicSpaceDyn(simdata, extThreshold=1e-4, fileout="antigenicspace.pdf")
#export table with dominance output
exportDominance(simdata, extThreshold=1e-4)
#export table with novelty output
exportNovelty(simdata, extThreshold=1e-4)

#export allelic matrix
exportAllelicMatrix(simdata, fileout="allelic.matrix.csv")

#set an extinction threshold
extThreshold=1e-50

#export sequences of dominant strains in time
exportSeqOfDomVariant(simdata, extThreshold=extThreshold, fileout="sequences.dominant.time.csv")

##plot 	sequences of dominant strains in time	
plotSeqsOfDomVariant(simdata, epiStruc=epiStruc,
	extThreshold=extThreshold, fileout="allelesintime.pdf")	
	
##testing differences in transmission phenotype
##(one strain transmits more than others)

	#define epitope structure, with 2 loci
	epiStruc<- c(2,2) 
	#get number of strains in the system
	nStrains<- extractNumberStrains(epiStruc) 
	#time at which the solver will stop
	tMax<- 1000
	#solver step
	tInt<- 0.005 
	#record solution every tObsPer steps
	tObsPer<- 100 
	#random initial conditions for infected
	initCondY <- runif(nStrains, 0, 1)*1e-4 

	#parameter values here defined by year
	#thus output time scale will be years
	beta<-  292
	sigma<-  (1/5)*365 #5 days infectious period
	mu<-  1/50 #50 years of life-span

	beta<- rep(292, nStrains)
	beta[2]<- 292*2

	#running model with weak cross-immunity
	gamma<- 0.1
	simdata<- runMANTIS(epiStruc, tMax, tObsPer, tInt, initCondY, beta, gamma, sigma, mu)

	#plotting output with weak cross-immunity to the screen
	plotY(simdata, xiObs=0.2, xfObs=0.5)	

##running a simulation in NSS with seasonality
##all transmission rates are affected equally

	#define antigenic structure
	epiStruc<- c(2,2)
	#time at which the solver will stop
	tMax<- 750
	#solver step
	tInt<- 0.01

	#record solution every tObsPer steps
	tObsPer<- 10

	#parameter values here defined by year
	#thus output time scale will be years
	beta<- 292
	sigma<- (1/5)*365 #5 days infectious period
	mu<- 1/50 #50 years of life-span

	#running model with weak cross-immunity
	gamma<- 0.01
	epsilon<- 0.25
	nLocus<- length(epiStruc)
	nStrains<- extractNumberStrains(epiStruc)
	infInitCond <- runif(nStrains, 0, 1)*1e-4

	simdata<- runMANTIS(epiStruc, tMax, tObsPer, tInt, infInitCond, beta, gamma, sigma, mu, epsilon)
	plotPathogenPrevalence(simdata, xiObs=0.8, xfObs=0.83)

##running a simulation in which strain 2 evolves
##an advantageous transmission phenotype at t=650	

	#define epitope structure, with 2 loci
	epiStruc<- c(3,2) 
	#get number of strains in the system
	nStrains<- extractNumberStrains(epiStruc) 
	#maximum time for the simulation
	tMax<- 1000 
	#time step for the simulation
	tInt<- 0.005 
	#record solution every tObsPer steps
	tObsPer<- 100 

	#random initial conditions for infected
	initCondY <- rep(1e-4, nStrains) 
	#strain 1 set with a slightly higher prevalence
	initCondY[1]<- initCondY[1]+ 0.00001 

	#default transmission for all strains
	beta<- rep(292, nStrains) 
	#cross immunity for this simulation
	gamma<- 0.98 
	#5 days infectious period
	sigma<- (1/5)*365 
	#50 years of life-span
	mu<- 1/50 

	#define here the extra parameters for this experiment
	tBetaChange= 650 #time at which beta will be changed
	changedStrain= 2 #which strain to change beta for
	changedBeta= 292*10

	#run MANTIS
	simdata<- runInvasionWithOneBetaChange(epiStruc, tMax, tObsPer, tInt, initCondY,
			 beta, gamma, sigma, mu, tBetaChange, changedBeta, changedStrain) 

	#plot entire simulation
	plotY(simdata, xiObs=0.65, xfObs=0.99, ymax=0.001, addLegend=TRUE) 	

}
