library(tidyverse)
library(magrittr)
library(ggforce)

my_cols <- c("#016394", "#f74700", "#b6003b", "#005342", "#1b264f")

all_results <- list.files("../../Results/VMR_R0", full.names=TRUE) %>%
  lapply(function(filename) {
    read_csv(filename, col_names=c("strain", "population",
                                   "n_extrema_noround", "n_extrema_round12",
                                   "n_extrema_round10", "n_extrema_round8",
                                   "n_extrema_round6", "n_extrema_round4",
                                   "movement", "gamma", "sigma", "R0"),
             col_types="dddddddddddd", progress=FALSE) %>%
      pivot_longer(matches("round\\d*$"), names_to="rounding_regime", values_to="n_extrema", names_pattern="n_extrema_(\\w+?$)") %>%
      group_by(population, strain, movement, gamma, sigma, rounding_regime) %>% mutate(rep = 1:n()) %>% ungroup() %>%
      pivot_wider(names_from=population, values_from=c(R0, n_extrema))
  }) %>%
  bind_rows() %>%
  mutate(rounding_regime = factor(rounding_regime, levels=c("noround", "round12", "round10", "round8", "round6", "round4"))) %>%
  group_by(strain, movement, gamma, sigma, R0_1, R0_2, rounding_regime) %>%
  mutate(rep = 1:n()) %>%
  mutate_at(vars(starts_with("n_extrema")), list(dynamics = ~case_when(. == -1 ~ "extinct",
                                                                       . == 0 ~ "steady",
                                                                       . == 1 ~ "unconverged",
                                                                       . > 1  ~ "cycles"))) %>%
  group_by(movement, gamma, sigma, R0_1, R0_2, rounding_regime, rep) %>%
  summarise_at(vars(ends_with("dynamics")), . %>% sort() %>% table() %>% {str_c(., names(.), collapse="")}) %>%
  mutate_at(vars(ends_with("dynamics")), ~case_when(. == "4steady" ~ "steady",
                                                       . == "4cycles" ~ "cycles",
                                                       # str_detect(., "unconverged") ~ "unconverged",
                                                       . == "2extinct2steady" ~ "extinct",
                                                       TRUE ~ "unconverged")) %>%
  rename_at(vars(ends_with("dynamics")), . %>% str_match("(_\\d+)_(dynamics)") %>% {str_c(.[,3], .[,2])}) %>%
  ungroup()

## what dynamics are expected?
expected_dynamics <- all_results %>%
  select(-R0_2, -dynamics_2, -rep) %>%
  group_by(gamma, sigma, R0_1, rounding_regime) %>%
  filter(movement == min(movement)) %>%
  select(-movement) %>%
  distinct() %>%
  rename(expected_1 = dynamics_1)

plot_data <- all_results %>%
  left_join(expected_dynamics,
            by=c("gamma", "sigma", "R0_1", "rounding_regime")) %>%
  filter_at(vars(starts_with("expected"), starts_with("dynamics")),
            all_vars(is_in(., c("cycles", "steady")))) %>%
  left_join(expected_dynamics %>% rename(R0_2 = R0_1, expected_2 = expected_1),
            by=c("gamma", "sigma", "R0_2", "rounding_regime")) %>%
  filter(expected_1 != expected_2) %>%
  mutate(trt = str_glue("{expected_1} %->% {expected_2}"))

gammas <- levels(factor(plot_data$gamma))
sigmas <- levels(fct_inorder(factor(plot_data %>% arrange(gamma) %>% use_series(sigma))))
rnaughts <- plot_data %>% group_by(gamma, sigma) %>% distinct(R0_1, R0_2) %>%
  filter(R0_1 < R0_2) %>% unite("x", R0_1, R0_2, sep=", ") %>% use_series(x)

labs <- list(bquote(gamma==.(gammas[1])*.(";")~sigma==.(sigmas[1])*.(";")~tilde(R[0])==.("[")*.(rnaughts[1])*.("]")),
             bquote(gamma==.(gammas[2])*.(";")~sigma==.(sigmas[2])*.(";")~tilde(R[0])==.("[")*.(rnaughts[2])*.("]")),
             bquote(gamma==.(gammas[3])*.(";")~sigma==.(sigmas[3])*.(";")~tilde(R[0])==.("[")*.(rnaughts[3])*.("]")))

ggplot(plot_data %>% filter(rounding_regime == "round8")) +
  aes(x=movement, y=as.integer(factor(dynamics_2)) - 1,
      colour=str_c(gamma, sigma), shape=str_c(gamma, sigma)) +
  geom_line(stat="smooth", method="glm", method.args=list(family="binomial"),
            se=FALSE, size=1.25, show.legend=FALSE, n=500, na.rm=TRUE) +
  geom_point(aes(size=str_c(gamma, sigma)), stroke=1.25, alpha=0.75) +
  xlab("Movement rate (proportion of source population per unit time)") +
  ylab("Destination disease dynamics") +
  scale_colour_manual(values=my_cols, labels=labs) +
  scale_shape_manual(values=c(1,3,4), labels=labs) +
  scale_size_manual(values=c(3,2.5,2.5), labels=labs) +
  scale_x_log10(breaks=c(1e-9, 1e-7, 1e-5, 0.001, 0.1),
                labels=c("1e-9", "1e-7", "1e-5", "0.001", "0.1")) +
  scale_y_continuous(breaks=0:1, labels=c("cycles", "steady")) +
  # facet_grid(rounding_regime~trt, labeller=label_parsed) +
  facet_wrap(.~trt, labeller=label_parsed, ncol=1) +
  theme_bw() +
  theme(legend.title=element_blank(),
        legend.position="bottom")
ggsave(filename="variable-movement-rate.png", width=7, height=7)
